"""Синтаксический анализатор арифметических выражений."""

import typing as tp

from .error import Error


STATE_IDLE = 0
STATE_OPERAND = 1
OPERATORS = {
    "+": (
        lambda left, right:
            (NUM_TYPE(left) if left is not None else 0) + NUM_TYPE(right)
    ),
    "-": (
        lambda left, right:
            (NUM_TYPE(left) if left is not None else 0) - NUM_TYPE(right)
    ),
    "*": lambda left, right: NUM_TYPE(left) * NUM_TYPE(right),
    "/": lambda left, right: NUM_TYPE(left) / NUM_TYPE(right),
}
OPENING_BRACKET = "("
CLOSING_BRACKET = ")"
BREAKE_TOKEN_SYMBOLS = (
    list(OPERATORS.keys()) + [OPENING_BRACKET, CLOSING_BRACKET]
)
NUM_TYPE = int


class SyntaxError(Error):
    """Ошибка синтаксиса."""

    def __init__(self, pos: int, *args) -> None:
        self.pos = pos
        super().__init__(pos, f"Syntax error at {pos} position.", *args)


class UnclosedBracket(SyntaxError):
    """Не закрыта скобка."""

    def __init__(self, pos: int, *args) -> None:
        super().__init__(pos, "Unclosed bracket.", *args)


class ExtraClosingBracket(SyntaxError):
    """Лишняя закрывающая скобка."""

    def __init__(self, pos: int, *args) -> None:
        super().__init__(pos, "Extra closing bracket.", *args)


class Token(object):
    """Элемент выражения."""

    text: str
    pos: int

    def __init__(self, text: str, pos: int) -> None:
        super().__init__()
        self.text = text
        self.pos = pos

    @property
    def priority(self) -> int:
        """Приоритет элемента выражения."""
        if self.text in ["+", "-"]:
            return 1
        if self.text in ["*", "/"]:
            return 2
        return 0

    def eval(self, left: tp.Optional[int], right: tp.Optional[int]) -> int:
        """Вычисление элемента выражения."""
        if self.text in OPERATORS.keys():
            return OPERATORS[self.text](left, right)
        return int(self.text)

    def __str__(self) -> str:
        return self.text

    def __eq__(self, o: object) -> bool:
        if isinstance(o, Token):
            return self.text == o.text and self.pos == o.pos
        return False


def lexer(expression: str) -> tp.Generator[Token, None, None]:
    """Лексический анализатор."""
    pos = 0
    token = ""
    token_pos = 0
    brackets = []
    for symbol in expression:
        if symbol.isspace():
            if token:
                yield Token(token, token_pos)
                token = ""
        elif symbol in BREAKE_TOKEN_SYMBOLS:
            if symbol == OPENING_BRACKET:
                brackets.append(pos)
            elif symbol == CLOSING_BRACKET:
                if len(brackets) == 0:
                    raise ExtraClosingBracket(pos)
                del brackets[-1:]
            if token:
                yield Token(token, token_pos)
                token = ""
            yield Token(symbol, pos)
        elif symbol.isdigit():
            if len(token) == 0:
                token_pos = pos
            token += symbol
        else:
            raise SyntaxError(pos)
        pos += 1
    if token:
        yield Token(token, token_pos)
    if len(brackets) > 0:
        raise UnclosedBracket(brackets[len(brackets) - 1])
